// JW.java
// $Id: JW.java,v 1.30 2000/08/07 14:10:02 bmahe Exp $
// (c) COPYRIGHT MIT, INRIA and Keio, 1999.
// Please first read the full copyright statement in file COPYRIGHT.html
package org.w3c.jwput;

import java.io.*;
import java.net.*;
import java.util.*;

import javax.swing.ProgressMonitor;
import javax.swing.ProgressMonitorInputStream;

import org.w3c.tools.codec.*;
import org.w3c.www.protocol.http.*;
import org.w3c.www.mime.*;
import org.w3c.www.http.*;

import org.w3c.util.XMLProperties;

/**
 * JW allows the user to perform HTTP requests
 * @version $Revision: 1.30 $
 * @author  Benot Mah (bmahe@w3.org)
 */
public class JW implements RequestObserver, JWOptions {

    /**
     * USER_AGENT HEADER
     */    
    public static final String USER_AGENT = "Winie/1.0.5";

    /**
     * BANNER
     */    
    public static final String BANNER     = "Winie - HTTP/1.1 PUT tool";

    /**
     * Software version
     */    
    public static final String VERSION    = "1.0.5";

    /**
     * Latest distribution URL
     */
    public final static String WINIE_LATEST_ZIP  = 
          "http://jigsaw.w3.org/Winie/download/winie-latest.zip";

    /**
     * Documentation URL
     */
    public final static String WINIE_DOC_URL  = 
          "http://jigsaw.w3.org/Winie/help.html";

    /**
     * Winie URL
     */
    public final static String WINIE_URL  = 
          "http://jigsaw.w3.org/Winie/";

    /**
     * Default mime type
     */
    public static final String DEFAULT_MIME = "application/octet-stream";

    protected JWUI             ui        = new org.w3c.jwput.tui.JWTUI();

    protected HttpManager      manager   = null;

    protected ETagsDatabase    etagsDB   = null;

    protected boolean          verbose   = false;

    protected boolean          prompt    = false;

    protected boolean          quiet     = false;

    protected LineNumberReader in        = null;

    protected String           user      = null;
    
    protected String           password  = null;

    protected String           puser     = null;
    
    protected String           ppassword = null;

    protected Properties       jwprops   = null;

    protected Hashtable        uriInfos  = new Hashtable();

    protected boolean          done      = false;

    protected boolean          aborted   = false;

    protected Request          request   = null;

    protected boolean isAbortedOrDone() {
	return (done||aborted);
    }

    protected boolean isAborted() {
	return aborted;
    }

    protected boolean isDone() {
	return done;
    }

    protected void setAborted(boolean aborted) {
	this.aborted = aborted;
    }

    protected void setDone(boolean done) {
	this.done = done;
    }
    
    public void abort() {
	setAborted(true);
	if (! request.isInterrupted()) {
	    request.interruptRequest();
	}
    }

    protected ETagsDatabase getETagsDatabase() {
	if (etagsDB == null) {
	    Properties props   = manager.getProperties();
	    String     dbClass = props.getProperty(ETAGS_DB);
	    
	    try {
		if (dbClass != null) {
		    Class c = Class.forName(dbClass);
		    etagsDB = (ETagsDatabase)c.newInstance();
		}
	    } catch (Exception ex) {
		ui.displayError("Unable to instanciate : "+dbClass);
	    } finally {
		if (etagsDB == null)
		    etagsDB = new ETagManager();
	    }
	}
	return etagsDB;
    }

    public String getETag(File file, String url) {
	return getETagsDatabase().getETag(file, url);
    }

    protected void saveEtag(File file, String url, String etag) {
	getETagsDatabase().saveETag(file, url, etag);
    }

    protected void removeEtag(File file, String url) {
	getETagsDatabase().removeETag(file, url);
    }

    /**
     * Get the HttpManager.
     * @param props the properties
     * @return a HttpManager instance
     */
    public HttpManager getManager(Properties props) {
	HttpManager manager = HttpManager.getManager(props);
	manager.setGlobalHeader("User-Agent", USER_AGENT);
	Enumeration propnames = props.keys();
	int idx = HEADER_PREFIX.length();
	while (propnames.hasMoreElements()) {
	    String name = (String)propnames.nextElement();
	    if (name.startsWith(HEADER_PREFIX)) {
		String header = name.substring(idx);
		manager.setGlobalHeader(header, props.getProperty(name));
	    }
	}
	return manager;
    }

    /**
     * Request Observer
     */
    public void notifyProgress(RequestEvent event) {
	if (verbose && !quiet) {
	    if (event.code == RequestEvent.EVT_CONNECTED)
		ui.log("## Connected!");
	    else if (event.code == RequestEvent.EVT_EMITED)
		ui.log("## Request headers emited!");
	    else if (event.code == RequestEvent.EVT_REPLIED)
		ui.log("## Reply headers have been received.");
	    else if (event.code == RequestEvent.EVT_CLOSED) {
		String error[] = {"The target server has improperly",
				  "closed the connection."};
		ui.displayError(error, "Connection closed");
	    } else if (event.code == RequestEvent.EVT_UNREACHABLE)
		ui.log("## Unreachable");
	}
    }

    /**
     * Get the port number included in this URL.
     * @return an int
     */
    public static int getPort(URL url) {
	int port = url.getPort();
	return (port == -1) ? 80 : port;
    }

    /**
     * Display the error reply (including the body).
     * @param reply the HTTP reply
     */
    protected void displayErrorMessage(Reply reply) {
	displayErrorMessage(reply, "");
    }

    /**
     * Display the error reply (including the body).
     * @param reply the HTTP reply
     * @param msg the message to display
     */
    protected void displayErrorMessage(Reply reply, String msg) {
	String array[] = {
	    msg,
	    "Error "+reply.getStatus()+": "+reply.getReason()
	};
	    
	ui.displayError(array, "Error");
	if (verbose) {
	    try {
		InputStream in = reply.getInputStream();
		if (in != null) {
		    LineNumberReader lin = 
			new LineNumberReader(new InputStreamReader(in));
		    String line = null;
		    while( (line = lin.readLine()) != null)
			ui.errlog(line);
		    ui.sync();
		}
	    } catch (IOException ex) {}
	}
    }

    /**
     * Display the reply (including the body).
     * @param reply the HTTP reply
     */
    protected void displayMessage(Reply reply) {
	ui.displayMessage("Reply "+reply.getStatus()+": "+reply.getReason());
	if (verbose) {
	    try {
		InputStream in = reply.getInputStream();
		if (in != null) {
		    LineNumberReader lin = 
			new LineNumberReader(new InputStreamReader(in));
		    String line = null;
		    while( (line = lin.readLine()) != null)
			ui.log(line);
		    ui.sync();
		}
	    } catch (IOException ex) {}
	}
    }

    /**
     * Create a HTTP Request using the given headers.
     * @param headers a Hastable containing HTTP headers
     * @return a HTTP Request
     */
    protected Request createRequest(Hashtable headers) {
	Request request = manager.createRequest();
	request.setObserver(this);
	Enumeration keys = headers.keys();
	while (keys.hasMoreElements()) {
	    String header = (String) keys.nextElement();
	    String value  = (String) headers.get(header);
	    request.setValue(header, value);
	}
	return request;
    }

    /**
     * Should be called to purge the reply, otherwise the connection won't
     * be marked idle and won't be reused.
     * @param reply the reply to purge.
     */
    protected void skipBody(Reply reply) {
	try {
	    if (reply.hasInputStream()) {
		InputStream in = reply.getInputStream();
		byte buffer[] = new byte[256];
		try {
		    while (in.read(buffer) != -1);
		} catch (IOException ioex) {
		    try { in.close(); } catch (IOException ioex2) {}
		}
	    }
	} catch (IOException ex) {
	    //nothing to do
	}
    }

    /**
     * Perform a HTTP/1.1 Request, handle authorization and redirection
     * automaticaly.
     * @param method the HTTP Method (HEAD, GET, PUT, TRACE, ...)
     * @param url the destination url
     * @return a Reply instance
     */
    public Reply performRequest(String method, String url) {
	return performRequest(method, url, null, new Hashtable(0));
    }

    /**
     * Perform a HTTP/1.1 Request, handle authorization and redirection
     * automaticaly.
     * @param method the HTTP Method (HEAD, GET, PUT, TRACE, ...)
     * @param url the destination url
     * @param headers HTTP headers
     * @return a Reply instance
     */
    public Reply performRequest(String method, String url, Hashtable headers) {
	return performRequest(method, url, null, headers);
    }

    /**
     * Perform a HTTP/1.1 Request, handle authorization and redirection
     * automaticaly. In case of redirection, Reply.getLocation() will
     * allways return the target URL.
     * @param method the HTTP Method (HEAD, GET, PUT, TRACE, ...)
     * @param url the destination url
     * @param body an inputStream (null if there is no body in the request)
     * @param headers HTTP headers
     * @return a Reply instance
     */
    public Reply performRequest(String method, 
				String url, 
				InputStream body, 
				Hashtable headers) 
    {
	return performRequest(method, url, body, headers, true);
    }

    /**
     * Perform a HTTP/1.1 Request, handle authorization and redirection
     * automaticaly. In case of redirection, Reply.getLocation() will
     * allways return the target URL.
     * @param method the HTTP Method (HEAD, GET, PUT, TRACE, ...)
     * @param url the destination url
     * @param body an inputStream (null if there is no body in the request)
     * @param headers HTTP headers
     * @param follow follow redirections?
     * @return a Reply instance
     */
    public Reply performRequest(String method, 
				String url, 
				InputStream body, 
				Hashtable headers,
				boolean follow) 
    {
	ProgressMonitor pm = null;
	Reply reply = null; 
	int at      = -1;
	int colon   = url.indexOf(':');
	if ((at = url.indexOf('@',6)) != -1) {
	    String auth = url.substring(colon+3,at);
	    int    sep  = -1;
	    if ((auth.indexOf('/') == -1) &&
		((sep = auth.indexOf(':')) != -1)) {
		user     = auth.substring(0,sep);
		password = auth.substring(sep+1);
	    }
	    url = "http://"+url.substring(at+1);
	}
	URL hurl  = null;
	try {
	    hurl = new URL(url);
	} catch (MalformedURLException ex) {
	    throw new RuntimeException("Invalid url : "+url.toString());
	}
	this.request = createRequest(headers);
	request.setMethod(method);
	request.setURL(hurl);
	if (body != null) {
	    request.setOutputStream(body);
	    if (body instanceof ProgressMonitorInputStream) {
		ProgressMonitorInputStream pin =
		    (ProgressMonitorInputStream) body;
		pm = pin.getProgressMonitor();
	    }
	}
	//password?
	String[] sauth = (String[]) uriInfos.get(method + ' ' + url);
	if (sauth == null) {
	    sauth = (String[]) uriInfos.get(method + ' ' + hurl.getHost());
	    if (sauth == null) {
		sauth = (String[]) uriInfos.get(hurl.getHost());
	    }
	}
	if (sauth == null) {
	    if (user != null) {
		Base64Encoder encoder = new Base64Encoder(user+":"+password);
		HttpCredential credential = 
		    HttpFactory.makeCredential("Basic");
		credential.setAuthParameter("cookie", encoder.processString());
		request.setAuthorization(credential);
	    }
	} else {
	    // the user/passwd is known, use it.
	    Base64Encoder encoder = new Base64Encoder(sauth[1]+":"+sauth[2]);
	    HttpCredential credential = HttpFactory.makeCredential("Basic");
	    credential.setAuthParameter("cookie", encoder.processString());
	    request.setAuthorization(credential);	    
	}
	if (puser != null) {
	    Base64Encoder encoder = new Base64Encoder(puser+":"+ppassword);
	    HttpCredential credential = HttpFactory.makeCredential("Basic");
	    credential.setAuthParameter("cookie", encoder.processString());
	    request.setProxyAuthorization(credential);
	}
	if (verbose && !quiet) {
	    if (manager.usingProxy()) {
		ui.log("Using proxy "+
				   request.getProxy().getHost()+":"+
				   getPort(request.getProxy()));
		ui.log("Sending Proxy request...\n");
		ui.sync();
	    } else {
		ui.log("Direct connection to "+
				   hurl.getHost()+":"+
				   getPort(hurl));
		ui.log("Sending HTTP request...\n");
		ui.sync();
	    }
	}
	try {
	    setDone(false);
	    setAborted(false);
	    int     redirect = 0;
	    do {
		if (verbose && !quiet) {
		    ui.log("Client Request : \n");
		    request.dump(ui.getLogStream());
		    ui.sync();
		}
		reply = manager.runRequest(request);
		if (verbose && !quiet) {
		    ui.log("Server response : \n");
		    reply.dump(ui.getLogStream());
		    ui.sync();
		}
		if (reply == null) {
		    return null;
		}
		if ((reply.getStatus() == HTTP.UNAUTHORIZED) && prompt) {
		    String realm = null;
		    HttpChallenge challenge = reply.getWWWAuthenticate();
		    if (challenge != null) {
			String scheme = challenge.getScheme();
			if ("Basic".equals(scheme)) {
			    realm = challenge.getAuthParameter("realm");
			}
		    }
		    String auth[] = 
			ui.getAuthentication(realm, "Authorization Required",
					     (user == null));
		    if (auth == null) {
			setDone(true);
		    } else {
			// add this unknown user/passwd in the hastable
			uriInfos.put(method + ' ' + url, auth);
			uriInfos.put(method + ' ' + hurl.getHost(), auth);
			if (uriInfos.get(hurl.getHost()) == null) {
			    uriInfos.put(hurl.getHost(), auth);
			}

			user     = auth[1];
			password = auth[2];
			Base64Encoder encoder = 
			    new Base64Encoder(user+":"+password);
			HttpCredential credential = 
			    HttpFactory.makeCredential("Basic");
			credential.setAuthParameter("cookie",
						    encoder.processString());
			request.setAuthorization(credential);
		    }
		} else if ((reply.getStatus() == HTTP.PROXY_AUTH_REQUIRED) 
			   && prompt ) {
		    String realm = null;
		    HttpChallenge challenge = reply.getProxyAuthenticate();
		    if (challenge != null) {
			String scheme = challenge.getScheme();
			if ("Basic".equals(scheme)) {
			    realm = challenge.getAuthParameter("realm");
			}
		    }
		    String auth[] = 
			ui.getAuthentication(realm, 
					     "Proxy Authorization Required",
					     (puser == null));
		    if (auth == null) {
			setDone(true);
		    } else {
			puser     = auth[1];
			ppassword = auth[2];
			Base64Encoder encoder = 
			    new Base64Encoder(puser+":"+ppassword);
			HttpCredential credential = 
			    HttpFactory.makeCredential("Basic");
			credential.setAuthParameter("cookie",
						    encoder.processString());
			request.setProxyAuthorization(credential);
		    }
		} else if ((reply.getStatus() == HTTP.MOVED_PERMANENTLY) ||
			   (reply.getStatus() == HTTP.FOUND) ||
			   (reply.getStatus() == HTTP.TEMPORARY_REDIRECT)) {
		    //redirections
		    if (! follow) {
			setDone(true);
		    } else if (redirect > 4) {
			ui.displayError("Too many redirections, aborted.");
			setDone(true);
		    } else {
			try {
			    URL location = new URL(reply.getLocation());
			    request.setURL(location);
			    redirect++;
			    if (! quiet) {
				ui.log("Redirected to "+
				       location.toExternalForm()+
				       "\n");
				ui.sync();
			    }
			} catch (Exception ex) {
			    ui.displayError(ex.getMessage());
			    setDone(true);
			}
		    }
		} else {
		    if ((redirect > 0) && (reply.getLocation() == null))
			reply.setLocation(request.getURL());
		    setDone(true);
		}
	    } while (! isAbortedOrDone());

	    if (isAborted() && (pm != null)) {
		pm.close();
	    }
	} catch (HttpException ex) {
	    if ((pm != null) && (pm.isCanceled())) {
		ui.setStatus("Action Aborted!");
		return null;
	    } 
	    if (isAborted()) {
		if (pm != null) {
		    pm.close();
		}
		return null;
	    }
	    try { body.close(); } catch (IOException ioex) {}
	    ui.displayError(ex.getMessage());
	}
	return reply;
    }

    /**
     * Get the General Properties.
     * @return a Property set
     */
    public Properties getProperties() {
	return manager.getProperties();
    }

    /**
     * Get the User Properties.
     * @return a Property set
     */
    public Properties getUserProperties() {
	return jwprops;
    }

    /**
     * Update the user properties
     * @param props the new properties
     */
    public void updateProperties(Properties props) {
	initialize(props);
    }

    protected void initialize(Properties props) {
	this.jwprops = props;

	String verbs = props.getProperty(VERBOSE_P);
	this.verbose = 
	    (verbs == null) ? false : verbs.equalsIgnoreCase(TRUE_P);
	String quiets = props.getProperty(QUIET_P);
	this.quiet = 
	    (quiets == null) ? false : quiets.equalsIgnoreCase(TRUE_P);

	String prompts = props.getProperty(PROMPT_P);
	this.prompt = 
	    (prompts == null) ? false : prompts.equalsIgnoreCase(TRUE_P);

	user     = props.getProperty(USER_P);
	password = props.getProperty(PASSWORD_P);

	String proxyset = props.getProperty(PROXY_SET_P);
	if ((proxyset != null) && (proxyset.equals(TRUE_P))) {
	    String socksh = props.getProperty(SOCKS_HOST_P);
	    String socksp = props.getProperty(SOCKS_PORT_P);
	    if ((socksh != null) && (! socksh.equals(""))) {
		System.setProperty(SOCKS_HOST_P, socksh);
		if (socksp != null)
		    System.setProperty(SOCKS_PORT_P, socksp);
	    } else {
		Properties sprops = System.getProperties();
		sprops.remove(SOCKS_HOST_P);
		sprops.remove(SOCKS_PORT_P);
		System.setProperties(sprops);
	    }
	} else {
	    Properties sprops = System.getProperties();
	    sprops.remove(SOCKS_HOST_P);
	    sprops.remove(SOCKS_PORT_P);
	    System.setProperties(sprops);
	}
    }
    
    /**
     * Check if the directory containing this file exists, if not
     * create it.
     * @param file the file
     */
    public static void check(File file) {
	if (! file.exists()) {
	    File dirs = new File(file.getParent());
	    if (!dirs.exists() && !dirs.mkdirs()) {
		System.out.println("Unable to create directory :"+
				   file.getParent());
		System.out.println("This directory is mandatory, please"+
				   " check the permissions");
		System.exit(-1);
	    }
	}
    }    

    public static File getPathConfig() {
	File pathconfig = null;
	String homedir = System.getProperty(HOMEDIR_P);
	if (homedir == null) {
	    homedir = System.getProperty("user.home");
	    pathconfig = new File(homedir, WPUTDIR);
	} else {
	    pathconfig = new File(homedir);
	}
	if (! pathconfig.exists()) {
	    pathconfig.mkdirs();
	}
	return pathconfig;
    }

    /**
     * Get the preferences file
     * @return a File instance
     */
    public static File getPreferencesFile() {
	File file = new File(getPathConfig(), WPUTRC);
	check(file);
	return file;
    }

    public static XMLProperties getPreferences() 
	throws IOException, FileNotFoundException
    {
	File file = new File(getPathConfig(), WPUTRC);
	XMLProperties props = new XMLProperties();
	props.load(file);
	return props;
    }
    
    public static void savePreferences(Properties prefs) 
	throws IOException
    {
	File file = getPreferencesFile();
	OutputStream out = 
	    new BufferedOutputStream(new FileOutputStream(file));
	prefs.store(out, "Winie User Preferences");
    }

    public static String getFileName(String path) {
	if ((path == null) || 
	    (path.length() == 0) ||
	    path.endsWith("/")) {
	    return null;
	}
	int idx = path.lastIndexOf('/');
	if (idx == -1) {
	    return path;
	}
	return path.substring(idx+1);
    }

}
